/* This file is part of the ili2ora project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.ehi.ili2ora.fromxtf;
import java.util.ArrayList;
import java.util.Vector;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.sql.PreparedStatement;
import java.sql.Connection;
import java.sql.Types;

import ch.ehi.basics.logging.EhiLogger;
import ch.ehi.ili2ora.fromili.TransferFromIli;
import ch.ehi.ili2ora.mapping.Mapping;

import ch.interlis.ili2c.metamodel.*;

import ch.interlis.iom.swig.iom_javaConstants;
import ch.interlis.iom.*;
import oracle.spatial.geometry.JGeometry;


/**
 * @author ce
 * @version $Revision: 1.0 $ $Date: 17.02.2005 $
 */
public class TransferFromXtf {
	private Mapping ili2sqlName=null;
	private HashMap tag2class=null;
	private HashSet unknownTypev=null;
	private TransferDescription td=null;
	private Connection conn=null;
	private java.sql.Timestamp today=null;
	private String dbusr=null;
	/** list of not yet processed struct values
	 */
	private ArrayList structQueue=null;
	public TransferFromXtf(Mapping ili2sqlName1,TransferDescription td1,Connection conn1,String dbusr1){
		ili2sqlName=ili2sqlName1;
		td=td1;
		conn=conn1;
		dbusr=dbusr1;
	}
		
	public void doit(IomFile iomFile){
		today=new java.sql.Timestamp(System.currentTimeMillis());
		tag2class=ch.interlis.ili2c.generator.XSDGenerator.getTagMap(td);
		unknownTypev=new HashSet();
		structQueue=new ArrayList();
		IomIterator iomBasketi=null;
		IomBasket iomBasket=null;
		boolean surfaceAsPolyline=true;
		iomBasketi=iomFile.iteratorbasket();
		// get first basket
		iomBasket=iomBasketi.nextbasket();
		// more baskets?
		while(iomBasket!=null){
			EhiLogger.logState("Basket "+iomBasket.getbaskettag()+"(oid "+iomBasket.getbasketoid()+")...");
			IomIterator iomObji=null;
			iomObji=iomBasket.iteratorobject();
		
			IomObject iomObj=null;
			iomObj=iomObji.nextobject();
			while(iomObj!=null){
				// translate object
				try{
					writeObject(iomObj,null);
				}catch(java.sql.SQLException ex){
					EhiLogger.logError("Object "+iomObj.getobjectoid()+" at (line "+iomObj.getobjectline()+",col "+iomObj.getobjectcol()+")",ex);
				}
				while(!structQueue.isEmpty()){
					StructWrapper struct=(StructWrapper)structQueue.remove(0); // get front
					try{
						writeObject(struct.getStruct(),struct);
					}catch(java.sql.SQLException ex){
						EhiLogger.logError("Object "+iomObj.getobjectoid()+"; Struct at (line "+struct.getStruct().getobjectline()+",col "+struct.getStruct().getobjectcol()+")",ex);
					}
				}
				iomObj=iomObji.nextobject();
			}
			iomBasket=iomBasketi.nextbasket();
		}
	}
	/** if structEle==null, iomObj is an object. If structEle!=null iomObj is a struct value.
	 */
	private void writeObject(IomObject iomObj,StructWrapper structEle)
		throws java.sql.SQLException
	{
		String tag=iomObj.getobjecttag();
		//EhiLogger.debug("tag "+tag);
		String sqlType=(String)ili2sqlName.mapIliClassName(tag);
		Viewable aclass=(Viewable)tag2class.get(tag);
		if(aclass==null){
			if(!unknownTypev.contains(tag)){
				EhiLogger.logError("unknown type <"+tag+">, line "+Integer.toString(iomObj.getobjectline())+", col "+Integer.toString(iomObj.getobjectcol()));
			}
			return;
		}
				
		 int sqlId;
		 if(structEle==null){
			sqlId=getObjSqlId(iomObj.getobjectoid());
		 }else{
		 	sqlId=newObjSqlId();
		 }
		 while(aclass!=null){
			String sqlname=(String)ili2sqlName.mapIliClassName(aclass.getScopedName(null));
			String insert = getInsertStmt(sqlname,aclass);
			PreparedStatement ps = conn.prepareStatement(insert);
			try{
				int valuei=1;
				ps.setInt(valuei, sqlId);
				valuei++;
				if(aclass.getExtending()==null){
					ps.setString(valuei, sqlType);
					valuei++;
					// if struct, add ref to parent
					if(structEle!=null){
						ps.setInt(valuei, structEle.getParentSqlId());
						valuei++;
						ps.setString(valuei, structEle.getParentSqlType());
						valuei++;
						// T_ParentAttr
						ps.setString(valuei, structEle.getParentSqlAttr());
						valuei++;
						// T_Seq
						ps.setInt(valuei, structEle.getStructi());
						valuei++;
					}
				}
			 
				Iterator iter = aclass.getDefinedAttributesAndRoles2();
				while (iter.hasNext()) {
					Viewable.TransferElement obj = (Viewable.TransferElement)iter.next();
					if (obj.obj instanceof AttributeDef) {
						AttributeDef attr = (AttributeDef) obj.obj;
						if(attr.getExtending()==null){
							 String attrName=attr.getName();
							Type type = attr.getDomain();
							if( (type instanceof TypeAlias) 
								&& TransferFromIli.isBoolean(td,type)) {
									String value=iomObj.getattrvalue(attrName);
									if(value!=null){
										if(value.equals("true")){
											ps.setString(valuei, "1");
										}else{
											ps.setString(valuei, "0");
										}
									}else{
										ps.setNull(valuei,Types.BIT);
									}
									valuei++;
							}else{
								type = attr.getDomainResolvingAliases();
							 
								if (type instanceof CompositionType){
									 // enqueue struct values
									 int structc=iomObj.getattrvaluecount(attrName);
									 for(int structi=0;structi<structc;structi++){
									 	IomObject struct=iomObj.getattrobj(attrName,structi);
									 	String sqlAttrName=ili2sqlName.mapIliAttributeDef(attr);
									 	enqueStructValue(sqlId,sqlType,sqlAttrName,struct,structi);
									 }
								}else if (type instanceof PolylineType){
									 IomObject value=iomObj.getattrobj(attrName,0);
									 if(value!=null){
										ps.setObject(valuei,getPolyline(value,false));
									 }else{
										 ps.setNull(valuei,Types.STRUCT,"MDSYS.SDO_GEOMETRY");
									 }
									 valuei++;
								 }else if(type instanceof SurfaceType){
									 IomObject value=iomObj.getattrobj(attrName,0);
									 if(value!=null){
										ps.setObject(valuei,getSurface(value,((SurfaceType)type).getLineAttributeStructure()!=null,false));
									 }else{
										 ps.setNull(valuei,Types.STRUCT,"MDSYS.SDO_GEOMETRY");
									 }
									 valuei++;
								 }else if(type instanceof AreaType){
									 IomObject value=iomObj.getattrobj(attrName,0);
									 if(value!=null){
										ps.setObject(valuei,getSurface(value,((AreaType)type).getLineAttributeStructure()!=null,false));
									 }else{
										 ps.setNull(valuei,Types.STRUCT,"MDSYS.SDO_GEOMETRY");
									 }
									 valuei++;
								 }else if(type instanceof CoordType){
									 IomObject value=iomObj.getattrobj(attrName,0);
									 if(value!=null){
										ps.setObject(valuei,getCoord(value));
									 }else{
										 ps.setNull(valuei,Types.STRUCT,"MDSYS.SDO_GEOMETRY");
									 }
									 valuei++;
								}else{
									String value=iomObj.getattrvalue(attrName);
									if(value!=null){
										ps.setString(valuei, value);
									}else{
										ps.setNull(valuei,Types.VARCHAR);
									}
									valuei++;
								}
							}
						}
					}
					if(obj.obj instanceof RoleDef){
						RoleDef role = (RoleDef) obj.obj;
						if(role.getExtending()==null){
							String roleName=role.getName();
							// a role of an embedded association?
							if(obj.embedded){
								AssociationDef roleOwner = (AssociationDef) role.getContainer();
								if(roleOwner.getDerivedFrom()==null){
									// not just a link?
									 IomObject structvalue=iomObj.getattrobj(roleName,0);
									if (roleOwner.getAttributes().hasNext()
										|| roleOwner.getLightweightAssociations().iterator().hasNext()) {
										 // TODO handle attributes of link
									}
									if(structvalue!=null){
										String refoid=structvalue.getobjectrefoid();
										long orderPos=structvalue.getobjectreforderpos();
										if(orderPos!=0){
										   // refoid,orderPos
										   //ret.setStringAttribute(roleName, refoid);
										   //ret.setStringAttribute(roleName+".orderPos", Long.toString(orderPos));
										}else{
										   // refoid
										   //ret.setStringAttribute(roleName, refoid);
										}
									   int refsqlId=getObjSqlId(refoid);
									   ps.setInt(valuei, refsqlId);
									}else{
										ps.setNull(valuei, Types.INTEGER);
									}
									valuei++;
								}
							 }else{
								 IomObject structvalue=iomObj.getattrobj(roleName,0);
								 String refoid=structvalue.getobjectrefoid();
								 long orderPos=structvalue.getobjectreforderpos();
								 if(orderPos!=0){
									// refoid,orderPos
									//ret.setStringAttribute(roleName, refoid);
									//ret.setStringAttribute(roleName+".orderPos", Long.toString(orderPos));
								 }else{
									// refoid
									//ret.setStringAttribute(roleName, refoid);
								 }
								int refsqlId=getObjSqlId(refoid);
								ps.setInt(valuei, refsqlId);
								valuei++;
							 }
						}
					 }
				}
				// T_LastChange
				ps.setTimestamp(valuei, today);
				valuei++;
				// T_CreateDate
				ps.setTimestamp(valuei, today);
				valuei++;
				// T_User
				ps.setString(valuei, dbusr);
				valuei++;
				ps.executeUpdate();
			}finally{
				ps.close();
			}
			aclass=(Viewable)aclass.getExtending();
		 }
	}
	private void enqueStructValue(int parentSqlId,String parentSqlType,String parentSqlAttr,IomObject struct,int structi)
	{
		structQueue.add(new StructWrapper(parentSqlId,parentSqlType,parentSqlAttr,struct,structi));
	}

	private java.sql.Struct getCoord(IomObject value)
		throws java.sql.SQLException
	{
		if(value!=null){
			
			String c1=value.getattrvalue("C1");
			String c2=value.getattrvalue("C2");
			String c3=value.getattrvalue("C3");
			double ordArray[]={0,0,0};
			ordArray[0] = Double.parseDouble(c1);
			ordArray[1] = Double.parseDouble(c2);
			JGeometry geom;
			if(c3==null){
				geom = JGeometry.createPoint(ordArray, 2,0);
			}else{
				ordArray[2] = Double.parseDouble(c3);
				geom = JGeometry.createPoint(ordArray, 3,0);
			}
			return JGeometry.store(geom,conn);
		}
		return null;
	}
	private void addPolyline(
		ArrayList elemInfo
		,ArrayList ordinates
		,IomObject obj
		,boolean hasLineAttr
		,boolean is3D
		)
		throws java.sql.SQLException
	{
		is3D=false;
		if(obj!=null){
			final int LINESTRING=2;
			final int STRAIGHT=1;
			final int ARC=2;
			int lastSegment=0;
			if(hasLineAttr){
				IomObject lineattr=obj.getattrobj("lineattr",0);
				if(lineattr!=null){
					//writeAttrs(out,lineattr);
					EhiLogger.logAdaption("Lineattributes not supported by Oracle; ignored");							
				}
			}
			boolean clipped=obj.getobjectconsistency()==iom_javaConstants.IOM_INCOMPLETE;
			for(int sequencei=0;sequencei<obj.getattrvaluecount("sequence");sequencei++){
				if(clipped){
					//out.startElement(tags::get_CLIPPED(),0,0);
				}else{
					// an unclipped polyline should have only one sequence element
					if(sequencei>0){
						EhiLogger.logError("unclipped polyline with multi 'sequence' elements");
						break;
					}
				}
				IomObject sequence=obj.getattrobj("sequence",sequencei);
				for(int segmenti=0;segmenti<sequence.getattrvaluecount("segment");segmenti++){
					IomObject segment=sequence.getattrobj("segment",segmenti);
					if(segment.getobjecttag().equals("COORD")){
						// COORD
						if(lastSegment!=STRAIGHT){
							addElemInfo(elemInfo,ordinates.size(),LINESTRING,STRAIGHT);
						}
						String c1=segment.getattrvalue("C1");
						String c2=segment.getattrvalue("C2");
						String c3=segment.getattrvalue("C3");
						ordinates.add(Double.valueOf(c1));
						ordinates.add(Double.valueOf(c2));
						if(is3D){
							ordinates.add(Double.valueOf(c3));
						}

					}else if(segment.getobjecttag().equals("ARC")){
						// ARC
						if(lastSegment!=ARC){
							addElemInfo(elemInfo,ordinates.size(),LINESTRING,ARC);
						}
						String c1=segment.getattrvalue("C1");
						String c2=segment.getattrvalue("C2");
						String c3=segment.getattrvalue("C3");
						String a1=segment.getattrvalue("A1");
						String a2=segment.getattrvalue("A2");
						ordinates.add(Double.valueOf(a1));
						ordinates.add(Double.valueOf(a2));
						if(is3D){
							// TODO calculate a3
							throw new IllegalStateException("TODO: calculate a3");
							//ordinates.add(Double.valueOf(a3));
						}
						ordinates.add(Double.valueOf(c1));
						ordinates.add(Double.valueOf(c2));
						if(is3D){
							ordinates.add(Double.valueOf(c3));
						}
					}else{
						// custum line form
						EhiLogger.logAdaption("custom line form not supported by Oracle; ignored");
						//out.startElement(segment->getTag(),0,0);
						//writeAttrs(out,segment);
						//out.endElement(/*segment*/);
					}

				}
				if(clipped){
					//out.endElement(/*CLIPPED*/);
				}
			}
		}
	}
	private java.sql.Struct getPolyline(IomObject obj,boolean is3D)
		throws java.sql.SQLException
	{
		is3D=false;
		if(obj!=null){
			ArrayList elemInfo=new ArrayList();
			ArrayList ordinates=new ArrayList();
			final int COMPOUND_LINESTRING=4;
			int eleInfoStartPos=elemInfo.size();
			addElemInfo(elemInfo,ordinates.size(),COMPOUND_LINESTRING,0);
			
			addPolyline(elemInfo,ordinates,obj,false,is3D);
			
			// patch counter in head triplet
			int elemInfoCount=(elemInfo.size()-eleInfoStartPos)/3-1;
			elemInfo.set(eleInfoStartPos+2,new Integer(elemInfoCount));
			int elemInfov[]=new int[elemInfo.size()];
			for(int i=0;i<elemInfov.length;i++){
				elemInfov[i]=((Integer)elemInfo.get(i)).intValue();
			}
			double ordinatesv[]=new double[ordinates.size()];
			for(int i=0;i<ordinatesv.length;i++){
				ordinatesv[i]=((Double)ordinates.get(i)).doubleValue();
			}
			JGeometry geom=new JGeometry(JGeometry.GTYPE_MULTICURVE,0,elemInfov,ordinatesv);
			return JGeometry.store(geom,conn);
		}
		return null;
	}
	private java.sql.Struct getSurface(IomObject obj,boolean hasLineAttr,boolean is3D)
		throws java.sql.SQLException
	{
		is3D=false;
		if(obj!=null){
			ArrayList elemInfo=new ArrayList();
			ArrayList ordinates=new ArrayList();
			final int EXTERIOR_POLYGON_RING=1005; // (must be specified in counterclockwise order)
			final int INTERIOR_POLYGON_RING=2005; // (must be specified in clockwise order)


			boolean clipped=obj.getobjectconsistency()==iom_javaConstants.IOM_INCOMPLETE;
			for(int surfacei=0;surfacei<obj.getattrvaluecount("surface");surfacei++){
				if(clipped){
					//out.startElement("CLIPPED",0,0);
				}else{
					// an unclipped surface should have only one surface element
					if(surfacei>0){
						EhiLogger.logError("unclipped surface with multi 'surface' elements");
						break;
					}
				}
				IomObject surface=obj.getattrobj("surface",surfacei);
				for(int boundaryi=0;boundaryi<surface.getattrvaluecount("boundary");boundaryi++){
					int eleInfoStartPos=elemInfo.size();
					if(boundaryi==0){
						addElemInfo(elemInfo,ordinates.size(),EXTERIOR_POLYGON_RING,0);
					}else{
						addElemInfo(elemInfo,ordinates.size(),INTERIOR_POLYGON_RING,0);
					}
					IomObject boundary=surface.getattrobj("boundary",boundaryi);
					for(int polylinei=0;polylinei<boundary.getattrvaluecount("polyline");polylinei++){
						IomObject polyline=boundary.getattrobj("polyline",polylinei);
						addPolyline(elemInfo,ordinates,polyline,true,is3D);
					}
					// patch counter in head triplet
					int elemInfoCount=(elemInfo.size()-eleInfoStartPos)/3-1;
					elemInfo.set(eleInfoStartPos+2,new Integer(elemInfoCount));
				}
				if(clipped){
					//out.endElement(/*CLIPPED*/);
				}
			}
			int elemInfov[]=new int[elemInfo.size()];
			for(int i=0;i<elemInfov.length;i++){
				elemInfov[i]=((Integer)elemInfo.get(i)).intValue();
			}
			double ordinatesv[]=new double[ordinates.size()];
			for(int i=0;i<ordinatesv.length;i++){
				ordinatesv[i]=((Double)ordinates.get(i)).doubleValue();
			}
			JGeometry geom=new JGeometry(JGeometry.GTYPE_POLYGON,0,elemInfov,ordinatesv);
			return JGeometry.store(geom,conn);
		}
		return null;
	}
	/** helper to build SDO input
	 */
	private void addElemInfo(ArrayList elemInfo,int sdoStartingOffset,int sdoEtype,int sdoInterpretation){
		elemInfo.add(new Integer(sdoStartingOffset+1));
		elemInfo.add(new Integer(sdoEtype));
		elemInfo.add(new Integer(sdoInterpretation));
	}
	private int sqlIdGen=1;
	private HashMap xtfId2sqlId=new HashMap();
	/** maps the xtfId to a sqlId.
	 */
	private int getObjSqlId(String xtfId){
		if(xtfId2sqlId.containsKey(xtfId)){
			return ((Integer)xtfId2sqlId.get(xtfId)).intValue();
		}
		int ret=sqlIdGen++;
		xtfId2sqlId.put(xtfId,new Integer(ret));
		return ret;
	}
	/** gets a new obj id.
	 */
	private int newObjSqlId(){
		int ret=sqlIdGen++;
		return ret;
	}
	/** creates an insert statement for a given viewable.
	 * @param sqlname table name of viewable
	 * @param aclass viewable
	 * @return insert statement
	 */
	private String createInsertStmt(String sqlname,Viewable aclass){
		StringBuffer ret = new StringBuffer();
		ret.append("INSERT INTO ");
		ret.append(sqlname);
		String sep=" (";
		Iterator iter = aclass.getDefinedAttributesAndRoles2();
		int valuei=0;
		// add T_Id
		ret.append(sep);
		sep=",";
		ret.append("T_Id");
		valuei++;
		// if root, add type
		if(aclass.getExtending()==null){
			ret.append(sep);
			sep=",";
			ret.append("T_Type");
			valuei++;
			// if STRUCTURE, add ref to parent
			if((aclass instanceof Table) && !((Table)aclass).isIdentifiable()){
				ret.append(sep);
				sep=",";
				ret.append("T_ParentId");
				valuei++;
				ret.append(sep);
				sep=",";
				ret.append("T_ParentType");
				valuei++;
				// attribute name in parent class
				ret.append(sep);
				sep=",";
				ret.append("T_ParentAttr");
				valuei++;
				// seqeunce (not null if LIST)
				ret.append(sep);
				sep=",";
				ret.append("T_Seq");
				valuei++;
			}
		}
		while (iter.hasNext()) {
		   Viewable.TransferElement obj = (Viewable.TransferElement)iter.next();
		   if (obj.obj instanceof AttributeDef) {
			   AttributeDef attr = (AttributeDef) obj.obj;
			   if(attr.getExtending()==null){
				Type type = attr.getDomainResolvingAliases();
				String attrName=ili2sqlName.mapIliAttributeDef(attr);
				if (type instanceof CompositionType){
				}else if (type instanceof PolylineType){
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
					 valuei++;
				 }else if(type instanceof SurfaceType){
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
					 valuei++;
				 }else if(type instanceof AreaType){
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
					 valuei++;
				 }else if(type instanceof CoordType){
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
					 valuei++;
				}else{
					 ret.append(sep);
					 sep=",";
					 ret.append(attrName);
					 valuei++;
				}
			   }
		   }
		   if(obj.obj instanceof RoleDef){
			   RoleDef role = (RoleDef) obj.obj;
			   if(role.getExtending()==null){
				String roleName=ili2sqlName.mapIliRoleDef(role);
				// a role of an embedded association?
				if(obj.embedded){
					AssociationDef roleOwner = (AssociationDef) role.getContainer();
					if(roleOwner.getDerivedFrom()==null){
						 // TODO if(orderPos!=0){
						 ret.append(sep);
						 sep=",";
						 ret.append(roleName);
						 valuei++;
					}
				 }else{
					 // TODO if(orderPos!=0){
					 ret.append(sep);
					 sep=",";
					 ret.append(roleName);
					 valuei++;
				 }
			   }
			}
		}
		// stdcols
		ret.append(sep);
		sep=",";
		ret.append("T_LastChange");
		valuei++;
		ret.append(sep);
		sep=",";
		ret.append("T_CreateDate");
		valuei++;
		ret.append(sep);
		sep=",";
		ret.append("T_User");
		valuei++;

		ret.append(") VALUES (");
		sep="";
		for(int i=0;i<valuei;i++){
			ret.append(sep);
			ret.append("?");
			sep=",";
		}
		ret.append(")");
		return ret.toString();
	}
	private HashMap insertStmts=new HashMap();
	/** gets an insert statement for a given viewable. Creates only a new
	 *  statement if this is not yet seen sqlname.
	 * @param sqlname table name of viewable
	 * @param aclass viewable
	 * @return insert statement
	 */
	private String getInsertStmt(String sqlname,Viewable aclass){
		if(insertStmts.containsKey(sqlname)){
			return (String)insertStmts.get(sqlname);
		}
		String stmt=createInsertStmt(sqlname,aclass);
		EhiLogger.traceBackendCmd(stmt);
		insertStmts.put(sqlname,stmt);
		return stmt;
	}
}
