// Copyright (C) Oracle Corporation 1999.        All Rights Reserved.

package sample;

import java.sql.*;
import oracle.sql.*;
import oracle.jdbc.driver.*;
import oracle.spatial.geometry.JGeometry;

/**
 * This sample program generates various geometries, converts them to
 * the SDO format and writes them to a database table.
 */
public class SampleNewToSDO
{
  private static String m_host = "eisenhut-c1d9hl.office.ehi.ch";
  private static String m_port = "1521";
  private static String m_sid = "md";
  private static String m_user = "cetest";
  private static String m_password = "cetest";

  private static String m_tableName;// = "TEST_TABLE3";
  private static String geomMetaDataTable = "user_sdo_geom_metadata";

  public static void main(String args[]) throws Exception
  {
    if (args.length != 1)
    {
      System.out.println("Parameters:");
      System.out.println("<Table name>: Table name for the result");
      return;
    }

    m_tableName = args[0].toUpperCase();
    String url = "jdbc:oracle:thin:@" + m_host + ":" + m_port + ":" + m_sid;
    OracleConnection conn = null;
    try
    {
      DriverManager.registerDriver(new OracleDriver());
      conn = (OracleConnection)DriverManager.getConnection(url, m_user, m_password);
    }
    catch(Exception e) { e.printStackTrace(); }

    prepareTableForData(conn);

    JGeometry geom;
    STRUCT str;
    // Set spatial reference system in which the geometries will be created
    //SRManager srManager = OraSpatialManager.getSpatialReferenceManager(conn);
    //SpatialReference sref = srManager.retrieve(8265);   // SRID = 8265: LAT-LON (NAD 83)
    //gF.setSpatialReference(sref);
    
    String update = "INSERT INTO " + m_tableName + " VALUES(?, ?)";
    PreparedStatement ps = conn.prepareStatement(update);

    //GeometryAdapter sdoAdapter = OraSpatialManager.getGeometryAdapter("SDO", "8.1.6",
    //                               null, STRUCT.class, null, conn);

    // create all types of different geometries
    {
      // create point
      System.out.println(" ");
      System.out.println("Create point...");
	  double ordArray[] = {11, 22};
      geom = JGeometry.createPoint(ordArray, 2,0);
      System.out.println("export to SDO...");
      str = (STRUCT) JGeometry.store(geom,conn);
      System.out.println("write to DB...");
      ps.setInt(1, 1);
      ps.setObject(2, str);
      ps.executeUpdate();
    }


    System.out.println("close connection...");
    ps.close();
    conn.close();
    System.out.println("Done with writing.");

    ////////////////////////////////////////////////////////////////////////////
    System.out.println("Now check DB...");
    System.out.println("////////////////////////////////");
    try
    {
      conn = (OracleConnection)DriverManager.getConnection(url, m_user, m_password);
    }
    catch(Exception e) { e.printStackTrace(); }
    conn.setDefaultRowPrefetch(100);

    String query = "SELECT * FROM " + m_tableName;
    Statement stm = conn.createStatement();
    OracleResultSet ors = (OracleResultSet)stm.executeQuery(query);
    while (ors.next())
    {
      JGeometry geometry = JGeometry.load((STRUCT)ors.getObject(2));
      System.out.println("Geometry type: " + geometry.getType());
    }
    stm.close();
    conn.close();
  }

  protected static void prepareTableForData(Connection conn)
  {
    ////////////////////////////////////////////////////////////////////////////
    // Preparation of the database
    ////////////////////////////////////////////////////////////////////////////

    // drop table
    System.out.println("Dropping old table...");
    Statement stmt = null;
    String update;
    try
    {
      stmt = conn.createStatement();
      update = "DROP TABLE " + m_tableName;
      stmt.executeUpdate(update);
      stmt.close();
    }
    catch (SQLException de) { System.out.println(de); }

    // delete reference to it from meta data table
    try
    {
      stmt = conn.createStatement();
      update =
        "DELETE FROM " + geomMetaDataTable + " WHERE table_name = '" + m_tableName + "'";
      stmt.executeUpdate(update);
      stmt.close();
    }
    catch (SQLException de) { System.out.println(de); }

    try
    {
      // create table
      System.out.println("Creating new table...");
      stmt = conn.createStatement();
      update = "CREATE TABLE " + m_tableName +
               " (gid NUMBER, geometry MDSYS.SDO_GEOMETRY)";
      stmt.executeUpdate(update);
      stmt.close();
    }
    catch (SQLException de) { System.out.println(de); }

    try
    {
      // add reference to it to meta data table
      stmt = conn.createStatement();
      update = "INSERT INTO " + geomMetaDataTable + " VALUES ('" + m_tableName +
               "', 'GEOMETRY', MDSYS.SDO_DIM_ARRAY(" +
               "MDSYS.SDO_DIM_ELEMENT('X', -180, 180, 0.0000005), " +
               "MDSYS.SDO_DIM_ELEMENT('Y', -90, 90, 0.0000005)), 1)";
      stmt.executeUpdate(update);
      stmt.close();
    }
    catch (SQLException de) { System.out.println(de); }
  }
}